 /*
  * Sample disk driver, from the beginning.
  */
 
 /* #include <linux/config.h> tpb */
 #include <linux/module.h>
 #include <linux/moduleparam.h>
 #include <linux/init.h>
 
 #include <linux/sched.h>
 #include <linux/kernel.h>       /* printk() */
 #include <linux/slab.h>         /* kmalloc() */
 #include <linux/fs.h>           /* everything... */
 #include <linux/errno.h>        /* error codes */
 #include <linux/timer.h>
 #include <linux/types.h>        /* size_t */
 #include <linux/fcntl.h>        /* O_ACCMODE */
 #include <linux/hdreg.h>        /* HDIO_GETGEO */
 #include <linux/kdev_t.h>
 #include <linux/vmalloc.h>
 #include <linux/genhd.h>
 #include <linux/blkdev.h>
 #include <linux/buffer_head.h>  /* invalidate_bdev */
 #include <linux/bio.h>

 #include <linux/random.h>
 #include <linux/time.h>

#include <asm-i386/checksum.h> /* for checksum function */

 
 MODULE_LICENSE("Dual BSD/GPL");
 
 static int sbull_major = 0;
 module_param(sbull_major, int, 0);
 static int hardsect_size = 512;
 module_param(hardsect_size, int, 0);
// static int nsectors = 1024;     /* How big the drive is */
 static int nsectors = 2048;
module_param(nsectors, int, 0);
 static int ndevices = 1;
 module_param(ndevices, int, 0);
 
 /*
  * The different "request modes" we can use.
  */
 enum {
         RM_SIMPLE  = 0, /* The extra-simple request function */
         RM_FULL    = 1, /* The full-blown version */
         RM_NOQUEUE = 2, /* Use make_request */
 };
 static int request_mode = RM_SIMPLE;

 module_param(request_mode, int, 0);
 
 /*
  * Minor number and partition management.
  */
 #define SBULL_MINORS    16
 #define MINOR_SHIFT     4
 #define DEVNUM(kdevnum) (MINOR(kdev_t_to_nr(kdevnum)) >> MINOR_SHIFT
 
 /*
  * We can tweak our hardware sector size, but the kernel talks to us
  * in terms of small sectors, always.
  */
 #define KERNEL_SECTOR_SIZE      512

 
 /*
  * After this much idle time, the driver will simulate a media change.
  */
 #define INVALIDATE_DELAY        30*HZ
 
 /*
  * The internal representation of each disk.
  */

 
/*
  * The internal representation of our device.
  */
 struct sbull_dev {
         int size;                       /* Device size in sectors */
         u8 *data[3];                       /* The data array */
         short users;                    /* How many users */
         spinlock_t lock;                /* For mutual exclusion */
         struct request_queue *queue;    /* The device request queue */
         struct gendisk *gd;             /* The gendisk structure */

	int layout_row;		/* The number of strips in one stripe */
	int layout_column;	/* The number of physical disks */
	int parity_row;		/* The number of parity blocks in each stripe */
	int parity_column;	/* The number of data blocks */
	int data_per_stripe;	/* Number of data blocks per stripe*/
	int parity_per_stripe;	/* Number of parity blocks per stripe*/

	int raid_layout[4][4];	/* For storing raid layout on disk*/
	int raid_parity[4][12];	/* For storing raid's parity*/	
	char* temp[11];		/* For temporary storage*/
	char* tempChksm[3];	/* For temporary buffer (size = chksm_size) */

	int block_per_chksm; // -- number of blocks per checksum
	int chksm_func; // -- checksum function. 0: sum
	int chksm_size; // -- size of each checksum
	
	u8 *tempU8; // -- temporary u8 type storage	
 };
 
 static struct sbull_dev *Devices = NULL;

static int time_to_seconds ( struct timeval *tstart, struct timeval *tfinish ) {
    int t;

    //t = (tfinish->tv_sec - tstart->tv_sec) + (tfinish->tv_usec - tstart->tv_usec) / 1e6;
    t = tfinish->tv_usec - tstart->tv_usec;
    printk(KERN_NOTICE "interval: %f\n", (tfinish->tv_sec - tstart->tv_sec) + (tfinish->tv_usec - tstart->tv_usec));

    return t;
}

 static void get_disk_sector(struct sbull_dev *dev, unsigned long input_sector, unsigned long *disk, unsigned long *output_sector){
	int offset = input_sector%dev->data_per_stripe;
	if(offset==0) offset = dev->data_per_stripe;
	
	*disk = 0;

	int i=0, j=0, quit = 0, row_location = 0;
	
	for(i=0; i<dev->layout_row; i++){
		for(j=0;j<dev->layout_column;j++){
			if(dev->raid_layout[i][j] == 1) offset --;
			if(offset == 0){
				*disk = j;
				row_location = i;
				quit = 1;
				break;
			}
		}
		if(quit==1) break;		
	}	
	
	*output_sector = (input_sector/dev->data_per_stripe) * dev->layout_row + row_location;
	return;

 }


static int calc_chksm(struct sbull_dev *dev, unsigned long disk, unsigned long sector, char* buffer) {
	int i;
	u8 chksm_val_U8;
	char chksm_val;
 
	memset (dev->temp[10], 0, KERNEL_SECTOR_SIZE);
	memset (dev->tempU8, 0, KERNEL_SECTOR_SIZE);

        // -- for checksum recalculate sector #
        int sector1 = sector;
        int return_val = 1;
        sector = (sector1 / dev->block_per_chksm) * (dev->block_per_chksm + 1) + sector1 % dev->block_per_chksm;
        printk (KERN_NOTICE "Remap from sector %d to sector %d.\n", sector1, sector);
        if (sector > nsectors) {
                printk (KERN_NOTICE "Raw read: Beyond-end write (%ld )\n", sector);
                return -1;
         }
        //else if((disk + sector)%8 != 0){

        if(1){
                memcpy(dev->tempU8, dev->data[disk] + (sector * KERNEL_SECTOR_SIZE), KERNEL_SECTOR_SIZE);
                //printk(KERN_NOTICE "Raw Read suceeded.\n");
                return_val = 1;
        }else{
                memcpy(dev->tempU8, dev->data[disk] + (sector * KERNEL_SECTOR_SIZE), KERNEL_SECTOR_SIZE);
                printk(KERN_NOTICE "Raw Read failed.\n");
                return_val = 1;
        }


	if(1){
                memcpy(dev->temp[10], dev->data[disk] + (sector * KERNEL_SECTOR_SIZE), KERNEL_SECTOR_SIZE);
                //printk(KERN_NOTICE "Raw Read suceeded.\n");
                return_val = 1;
        }else{
                memcpy(dev->temp[10], dev->data[disk] + (sector * KERNEL_SECTOR_SIZE), KERNEL_SECTOR_SIZE);
                printk(KERN_NOTICE "Raw Read failed.\n");
                return_val = 1;
        }

	printk(KERN_NOTICE "data sector for chksm: %s", dev->temp[9]);	
	if (dev->chksm_func == 0) {
	// -- use mirroring as chksm
		//memcpy(buffer, dev->temp[10], KERNEL_SECTOR_SIZE);
		return 1;
	}
	if (dev->chksm_func == 1) {
	// -- use sum as chksm function
		for (i = 0; i < KERNEL_SECTOR_SIZE; i++) {
			//chksm_val = chksm_val + simple_strtol(dev->temp[10]+i, NULL, 10);
			chksm_val_U8 = chksm_val_U8 + dev->tempU8[i];
			chksm_val = chksm_val & dev->temp[10][i];
		}
		snprintf(&chksm_val, sizeof(char), "%d", chksm_val_U8);
		//buffer[0] = chksm_val;
		//buffer[0] = chksm_val_U8;
		printk(KERN_NOTICE "checksum value for disk %d sector %d: %d.\n", disk, sector, chksm_val_U8);
	}
	if (dev->chksm_func == 2) {
	// -- use checksum function from asm-i386
		csum_partial(dev->temp[10], KERNEL_SECTOR_SIZE, 0);
	}

	return return_val;
}

static int low_read_sector_chksm(struct sbull_dev *dev, unsigned long disk, unsigned long sector, char *buffer){

	// -- get time
/*
	struct timeval *sTime;
	struct timeval *eTime;
	double interval;
	do_gettimeofday(sTime);
*/

        // -- for checksum recalculate sector #
        int sector1 = sector;
        sector = (sector1 / dev->block_per_chksm) * (dev->block_per_chksm + 1) + dev->block_per_chksm;
	
        printk (KERN_NOTICE "Chksm sector for sector %d is sector %d.\n", sector1, sector);
        if (sector > nsectors) {
                printk (KERN_NOTICE "Raw read chksm: Beyond-end write (%ld )\n", sector);
                return 0;
         }


        if(1){
                memcpy(buffer, dev->data[disk] + (sector * KERNEL_SECTOR_SIZE) + (sector % dev->block_per_chksm) * dev->chksm_size, dev->chksm_size);
		//do_gettimeofday(eTime);
		//interval = time_to_seconds(sTime, eTime);
                //printk(KERN_NOTICE "Raw Read Chksm suceeded. Time for reading: %f\n", interval);
		printk(KERN_NOTICE "Raw Read Chksm suceeded.n");
                return 1;
        }else{
                memcpy(buffer, dev->data[disk] + (sector * KERNEL_SECTOR_SIZE) + (sector % dev->block_per_chksm) * dev->chksm_size, dev->chksm_size);
                //do_gettimeofday(eTime);
                //interval = time_to_seconds(sTime, eTime);
                //printk(KERN_NOTICE "Raw Read Chksm failed. Time for reading: %f\n", interval);
		printk(KERN_NOTICE "Raw Read Chksm failed.n");
                return 1;
        }
}

 static int low_read_sector(struct sbull_dev *dev, unsigned long disk, unsigned long sector, char *buffer){

        // -- get time
/*
	struct timeval *sTime;
        struct timeval *eTime;
        double interval;
        do_gettimeofday(sTime);
*/	
	// -- for checksum recalculate sector #
	int sector1 = sector;
	int return_val = 1;
	sector = (sector1 / dev->block_per_chksm) * (dev->block_per_chksm + 1) + sector1 % dev->block_per_chksm;
	printk (KERN_NOTICE "Remap from sector %d to sector %d.\n", sector1, sector);
	if (sector > nsectors) {
                printk (KERN_NOTICE "Raw read: Beyond-end write (%ld )\n", sector);
                return 0;
        }


	if(1){
		memcpy(buffer, dev->data[disk] + (sector * KERNEL_SECTOR_SIZE), KERNEL_SECTOR_SIZE);
                printk(KERN_NOTICE "Raw Read suceeded.\n");
		return_val = 1;
	}else{
		memcpy(buffer, dev->data[disk] + (sector * KERNEL_SECTOR_SIZE), KERNEL_SECTOR_SIZE);
                printk(KERN_NOTICE "Raw Read failed.\n");
		return_val = 1;
	}
        memset(dev->tempChksm[0], 0, dev->chksm_size);
        memset(dev->tempChksm[1], 0, dev->chksm_size);
 
 
        int readChksm = low_read_sector_chksm(dev, disk, sector, dev->tempChksm[1]);
	if (readChksm != 1) {
		printk(KERN_NOTICE "error reading cheksum for sidk %d, sector %d.\n", disk, sector);
		return_val = -1;
	} 
        calc_chksm(dev, disk, sector, dev->tempChksm[0]);

        if (strcmp(dev->tempChksm[0], dev->tempChksm[1])) {
                printk(KERN_NOTICE "disk %d, sector %d diff from checksum.\n", disk, sector);
                return_val = -1;
        }
	
       //do_gettimeofday(eTime);
       //interval = time_to_seconds(sTime, eTime);
       //printk(KERN_NOTICE "Raw Read time (including reading checksum): %f\n", interval);
	// -- for inserting random errors
/*
	int myRand;
	get_random_bytes(&myRand, sizeof(int));
	myRand = myRand % 5;
	if (myRand == 0) {
		return_val = -1;
	}
	else if (myRand == 1) {
		return_val = 0;
	}
	else {
		return_val = 1;
	}
*/
	return return_val;

 }

static char* parity(struct sbull_dev *dev, char *input[], int length){
	memset(dev->temp[9],0,KERNEL_SECTOR_SIZE);
	int i, j;
	for(i=0;i<KERNEL_SECTOR_SIZE;i++){
		char t = *(input[0] + i);
		for(j=1;j<length;j++){
			t = t ^ *(input[j] + i);
		}
		memcpy(dev->temp[9] + i, &t, sizeof(char));
	}
	return dev->temp[9];
}


 static char* read_sector(struct sbull_dev *dev, unsigned long disk, unsigned long sector){



	memset(dev->temp[9],0,KERNEL_SECTOR_SIZE);

	int read = low_read_sector(dev, disk, sector, dev->temp[9]);

	int myRand;
        get_random_bytes(&myRand, sizeof(int));
        myRand = myRand % 5;
        if (myRand == 0) {
               	read = -1;
        }
        else if (myRand == 1) {
               	read = 0;
        }
        else {
               	read = 1;
        }
		
	if(read==0 || read == -1){

		if (read == 0) {
			printk(KERN_NOTICE "No data read from disk %d, sector %d.\n", disk, sector);
		}
		else {
			printk(KERN_NOTICE "Data read not consistent with checksum. disk %d, sector %d.\n", disk, sector);
		}

		// Reconstruct the sector
		printk(KERN_NOTICE  "start reconstruction\n");
		int disk_offset = sector%dev->layout_row;	// See which row in the layout_matrix the sector is

		// Determine which column in the parity natrix the sector is
		int data_position, i, j, temp;
		temp = 0;
		for(i=0; i<dev->layout_row; i++){
			for(j=0;j<dev->layout_column;j++){
				if(dev->raid_layout[i][j]==1)
					temp++;
				if(i==disk_offset && j==disk)
					data_position = temp-1;
			}
		}
		printk(KERN_NOTICE  "disk=%d, sector=%d, disk_offset=%d, data_position=%d\n",disk, sector, disk_offset, data_position);

		for(i=0; i<dev->parity_row; i++){
		
			if(dev->raid_parity[i][data_position]==1){
				printk(KERN_NOTICE  "Found a reconstruction candidate\n");
				// Found one candidate for reconstruction
				int term_number, con;
				con = 0;
				term_number = 0;
				// Get the number of data blocks involved in calculating parity
				for(j=0;j<dev->parity_column;j++){
					if(dev->raid_parity[i][j]==1){
						term_number++;						
					}
				}

				// Read blocks for reconstruction
				int k;
				for(k=0;k<9;k++)
					memset(dev->temp[k],0,KERNEL_SECTOR_SIZE);
				int index;
				index = 0;
				for(j=0;j<dev->parity_column;j++){
					if(dev->raid_parity[i][j]==1 && j!=data_position){
						int count,m,n,raw_disk,raw_sector,found;
						count = -1;
						found = 0;
						raw_disk = 0;
						raw_sector = 0;
						for(m=0;m<dev->layout_row;m++){
							for(n=0;n<dev->layout_column;n++){
								if(dev->raid_layout[m][n] == 1) count++;
								if(count==j){
									raw_disk = n;
									raw_sector = (sector/dev->layout_row)*dev->layout_row +m;
									found = 1;
									break;
								}
							}
							if(found==1) break;
						}
						memset(dev->temp[index],0,KERNEL_SECTOR_SIZE);
						printk(KERN_NOTICE "Reading a construction block at disk %d, sector %d\n",raw_disk, raw_sector);
						int read1 = low_read_sector(dev, raw_disk, raw_sector, dev->temp[index]);
																	
						index++;
						// If read fails, continue with the next candidate
						if(read1==0){
							con = 1;
							printk(KERN_NOTICE "Reading construction block failed\n");
							break;
						}
						else if (read1 == -1) {
							con = 1;
							printk(KERN_NOTICE "Block for reconstruction is inconsistent with checksum.\n");
							break;
						} 
						else {
							printk(KERN_NOTICE "Reading construction block suceeded\n");
						}
					}
				}

				if(con==1) continue;

				// Read the parity block
				int count,m,n,parity_disk,parity_sector,found;
				count = -1;
				found = 0;
				parity_disk = 0;
				parity_sector = 0;
				for(m=0;m<dev->layout_row;m++){
					for(n=0;n<dev->layout_column;n++){
						if(dev->raid_layout[m][n] == 0) count++;
						if(count==i){
							parity_disk = n;
							parity_sector = (sector/dev->layout_row)*dev->layout_row +m;
							found = 1;
							break;
						}
					}
					if(found==1) break;
				}
				
				memset(dev->temp[term_number-1],0,KERNEL_SECTOR_SIZE);
				
				printk(KERN_NOTICE "Reading the parity block at disk %d, sector %d\n", parity_disk, parity_sector);
				int read1 = low_read_sector(dev, parity_disk, parity_sector, dev->temp[term_number-1]);			

				// If read fails, continue with the next candidate
				if(read1==0){
					printk(KERN_NOTICE "Reading the parity block failed\n");		
					continue;
														
				}
                                else if (read1 == -1) {
                                	con = 1;
                                        printk(KERN_NOTICE "Block for reconstruction is inconsistent with checksum.\n");
                                        continue;
                                } 
 				else{
					memset(dev->temp[9],0,KERNEL_SECTOR_SIZE);
					dev->temp[9] = parity(dev, dev->temp, term_number);
					
					printk(KERN_NOTICE "Reading the parity block suceeded, Reconstruction succeeded\n");
					//int q;
					//for(q=0;q<KERNEL_SECTOR_SIZE;q++){
					//	if(*(buffer+q) != *(buffer1+q))
					//		printk("compare: %c, %c\n",(buffer+q), (buffer1+q));
					//}

					return dev->temp[9];
				}
			}
		}
		printk(KERN_NOTICE "Unable to reconstruct\n");
		// Reconstruction was unsuccessful
		return NULL;
		
		//return buffer;
	}
	else if(read==1) {
 		printk(KERN_NOTICE "read suceeded.\n");
		return dev->temp[9];
	}
	else 
		return NULL;
 }

 /*
  * Handle an I/O request.
  */
 static void sbull_transfer(struct sbull_dev *dev, unsigned long sector,
                 unsigned long nsect, char *buffer, int write)
 {
	
	int i=0;
	
        struct timeval sTime;
      	struct timeval eTime;
        int interval;

	for(i=0; i < nsect;i++){
		unsigned long disk = 0, output_sector = 0;
		get_disk_sector(dev, sector + i, &disk, &output_sector);

		unsigned long disk_offset = output_sector * KERNEL_SECTOR_SIZE;
		unsigned long buffer_offset = i * KERNEL_SECTOR_SIZE;

		if (output_sector > nsectors) {
                	printk (KERN_NOTICE  "Beyond-end write (%ld )\n", disk_offset);
                	return;
         	}

		if(write){
			printk(KERN_NOTICE "start writing disk %d, sector %d\n", disk, output_sector);
		
			printk(KERN_NOTICE "remap writing from sector %d to sector ", output_sector);

	        	do_gettimeofday(&sTime);


			int old_output_sector = output_sector;	
			output_sector = (output_sector / dev->block_per_chksm) * (dev->block_per_chksm + 1) + output_sector % dev->block_per_chksm;
			printk(KERN_NOTICE "%d\n", output_sector);
			disk_offset = output_sector * KERNEL_SECTOR_SIZE;
	                if (output_sector > nsectors) {
        	                printk (KERN_NOTICE  "Beyond-end write (%ld )\n", disk_offset);
                	        return;
                	}


			// read the old data block			
			memset(dev->temp[9], 0, KERNEL_SECTOR_SIZE);
			
			//dev->temp[0] = read_sector(dev, disk, output_sector);
			if(!dev->temp[9]){					
				printk ("Unable to read the old sector.\n");
				// write data
				memcpy(dev->data[disk] + disk_offset, buffer + buffer_offset, KERNEL_SECTOR_SIZE);
				return 0;
			}

			//char *old_sector;
			//old_sector = vmalloc(KERNEL_SECTOR_SIZE);
			//if (old_sector == NULL) {
			//	printk (KERN_NOTICE  "vmalloc failure.\n");
			//	return 0;
			//}
			//int read_r = low_read_sector(dev, disk, output_sector, old_sector);
			//if(read_r!=1){
			//	printk(KERN_NOTICE "Unable to read the old sector.\n");
			//	memcpy(dev->data[disk] + disk_offset, buffer + buffer_offset, KERNEL_SECTOR_SIZE);
			//	return 0;
			//}
			
			printk (KERN_NOTICE "Old sector read.\n");

			// write data
			
			memcpy(dev->data[disk] + disk_offset, buffer + buffer_offset, KERNEL_SECTOR_SIZE);

			// write checksum
			memset(dev->tempChksm[0], 0, dev->chksm_size);
			calc_chksm(dev, disk, old_output_sector, dev->tempChksm[0]);
			int chksm_offset = (output_sector - output_sector % (dev->block_per_chksm + 1) + dev->block_per_chksm) * KERNEL_SECTOR_SIZE + (output_sector % (dev->block_per_chksm + 1)) * dev->chksm_size;
			memcpy(dev->data[disk] + chksm_offset, dev->tempChksm[0], dev->chksm_size);
			printk(KERN_NOTICE "checksum for disk %d, sector %d, wrriten at sector %d, check sum size: %d\n", disk, output_sector, chksm_offset, dev->chksm_size);
			
			// write new parity
			int offset = (sector+i)%dev->data_per_stripe;
			int m = 0;
			for(m=0;m<dev->parity_row;m++){				
				if(dev->raid_parity[m][offset] == 1){
					// Find a parity block to update
					printk (KERN_NOTICE "Found one parity block to update.\n");

					// Read the old parity
					int parity_offset = m + 1;
					int quit = 0, parity_disk = 0, parity_row = 0, parity_sector = 0;
					int p=0, q=0;

					for(p=0;p<dev->layout_row;p++){
						for(q=0;q<dev->layout_column;q++){
							if (dev->raid_layout[p][q] == 0) parity_offset --;
							if(parity_offset == 0){
								parity_disk = q;
								parity_row = p;
								quit = 1;
								break;				
							}
						}
						if(quit==1) break;		
					}

					parity_sector = ((sector + i)/dev->data_per_stripe) * dev->layout_row + parity_row;
					printk (KERN_NOTICE "Reading the old parity on disk %d, sector %d.\n", parity_disk, parity_sector);
					

					memset(dev->temp[1],0,KERNEL_SECTOR_SIZE);
					dev->temp[1] = read_sector(dev, parity_disk, parity_sector);
					if(!dev->temp[1]){					
						printk (KERN_NOTICE "Unable to read the old parity.\n");
						return 0;
					}

					//char *old_parity;
					//old_parity = vmalloc(KERNEL_SECTOR_SIZE);
					//if (old_sector == NULL) {
					//	printk (KERN_NOTICE  "vmalloc failure.\n");
					//	return 0;
					//}
					//int read_r = low_read_sector(dev, parity_disk, parity_sector, old_parity);
					//if(read_r!=1){
					//	printk(KERN_NOTICE "Unable to read the old parity.\n");
					//	return 0;
					//}

					printk(KERN_NOTICE "Old parity read.\n");
					// Calculate the new parity
															
					memset(dev->temp[2], 0, KERNEL_SECTOR_SIZE);
					memcpy(dev->temp[2], buffer + buffer_offset, KERNEL_SECTOR_SIZE);
					
					dev->temp[9] = parity(dev, dev->temp, 3);


					// Write new parity
					int new_parity_sector = (parity_sector / dev->block_per_chksm) * (dev->block_per_chksm + 1) + parity_sector % dev->block_per_chksm;
					printk (KERN_NOTICE "Writing the new parity on sector %d.\n", new_parity_sector);
					memcpy(dev->data[parity_disk] + new_parity_sector * KERNEL_SECTOR_SIZE, dev->temp[9], KERNEL_SECTOR_SIZE);

		                        // write checksum for parity block
                		        memset(dev->tempChksm[1], 0, dev->chksm_size);
		                        calc_chksm(dev, disk, parity_sector, dev->tempChksm[1]);
                		        int parity_chksm_offset = (new_parity_sector - new_parity_sector % (dev->block_per_chksm + 1) + dev->block_per_chksm) * KERNEL_SECTOR_SIZE + (new_parity_sector % (dev->block_per_chksm + 1)) * dev->chksm_size;
                		        memcpy(dev->data[disk] + parity_chksm_offset, dev->tempChksm[1], dev->chksm_size);
                        		printk(KERN_NOTICE "checksum for parity: disk %l, sector %l, wrriten at sector %d", disk, parity_sector, parity_chksm_offset);



				}
			}

			do_gettimeofday(&eTime);
 	        	interval = time_to_seconds(&sTime, &eTime);
        	        printk(KERN_NOTICE "Writing time (including reading checksum): %d, %d, interval: %d\n", sTime.tv_sec, eTime.tv_sec, interval);
 
			printk(KERN_NOTICE "Writing completed\n");
		}
		else{
			// read
			printk(KERN_NOTICE "start reading disk %d, sector %d\n",disk, output_sector);

        		do_gettimeofday(&sTime);

			char *read_data = read_sector(dev, disk, output_sector);
			if(read_data!=NULL){
				memcpy(buffer + buffer_offset, read_data, KERNEL_SECTOR_SIZE);
			}
			else{
				printk (KERN_NOTICE  "Unable to read.\n");
				return;
			}
       			do_gettimeofday(&eTime);
       			interval = time_to_seconds(&sTime, &eTime);
       			printk(KERN_NOTICE "Read time (including reading checksum): %d, %d, interval: %d\n", sTime.tv_sec, eTime.tv_sec, interval);
			printk(KERN_NOTICE "reading completed\n");
		}
	}

	
         
         /*unsigned long nbytes = nsect*KERNEL_SECTOR_SIZE;
 
         

         if (write)
                 memcpy(dev->data + offset, buffer, nbytes);
         else
                 memcpy(buffer, dev->data + offset, nbytes);
	
	int i = 0, j = 0;
	for(i=0;i<dev->layout_row;i++){
		for(j=0;j<dev->layout_row;j++)
			printk("%d ",dev->raid_layout[i][j]);
		printk("\n");
	}*/
 }
 
 /*
  * The simple form of the request function.
  */
 static void sbull_request(request_queue_t *q)
 {
         struct request *req;
 
         while ((req = elv_next_request(q)) != NULL) {
                 struct sbull_dev *dev = req->rq_disk->private_data;
                 if (! blk_fs_request(req)) {
                         printk (KERN_NOTICE "Skip non-fs request\n");
                         end_request(req, 0);
                         continue;
                 }
     //          printk (KERN_NOTICE "Req dev %d dir %ld sec %ld, nr %d f %lx\n",
     //                          dev - Devices, rq_data_dir(req),
     //                          req->sector, req->current_nr_sectors,
     //                          req->flags);
                 sbull_transfer(dev, req->sector, req->current_nr_sectors,
                                 req->buffer, rq_data_dir(req));
                 end_request(req, 1);
         }
 }
 
 
 /*
  * Transfer a single BIO.
  */
 static int sbull_xfer_bio(struct sbull_dev *dev, struct bio *bio)
 {
         int i;
         struct bio_vec *bvec;
         sector_t sector = bio->bi_sector;
 
         /* Do each segment independently. */
         bio_for_each_segment(bvec, bio, i) {
                 char *buffer = __bio_kmap_atomic(bio, i, KM_USER0);
                 sbull_transfer(dev, sector, bio_cur_sectors(bio),
                                 buffer, bio_data_dir(bio) == WRITE);
                 sector += bio_cur_sectors(bio);
                 __bio_kunmap_atomic(bio, KM_USER0);
         }
         return 0; /* Always "succeed" */
 }
 
 /*
  * Transfer a full request.
  */
 static int sbull_xfer_request(struct sbull_dev *dev, struct request *req)
 {
         struct bio *bio;
         int nsect = 0;
     
         rq_for_each_bio(bio, req) {
                 sbull_xfer_bio(dev, bio);
                 nsect += bio->bi_size/KERNEL_SECTOR_SIZE;
         }
         return nsect;
 }
 
 
 
 /*
  * Smarter request function that "handles clustering".
  */
 static void sbull_full_request(request_queue_t *q)
 {
         struct request *req;
         int sectors_xferred;
         struct sbull_dev *dev = q->queuedata;
 
         while ((req = elv_next_request(q)) != NULL) {
                 if (! blk_fs_request(req)) {
                         printk (KERN_NOTICE "Skip non-fs request\n");
                         end_request(req, 0);
                         continue;
                 }
                 sectors_xferred = sbull_xfer_request(dev, req);
                 if (! end_that_request_first(req, 1, sectors_xferred)) {
                         blkdev_dequeue_request(req);
 /*                      end_that_request_last(req); tpb */
                         end_that_request_last(req, 0); /* tpb */
 /* FIX ME - tpb
    Is 0 the correct value her for boolean parameter uptodate?  tpb
    Are we holding the queue lock here? tpb */
                 }
         }
 }
 
 
 
 /*
  * The direct make request version.
  */
 static int sbull_make_request(request_queue_t *q, struct bio *bio)
 {
         struct sbull_dev *dev = q->queuedata;
         int status;
 
         status = sbull_xfer_bio(dev, bio);
         bio_endio(bio, bio->bi_size, status);
         return 0;
 }
 
 
 /*
  * Open and close.
  */
 
 static int sbull_open(struct inode *inode, struct file *filp)
 {
         struct sbull_dev *dev = inode->i_bdev->bd_disk->private_data;
 
         filp->private_data = dev;
         spin_lock(&dev->lock);
         if (! dev->users) 
                 check_disk_change(inode->i_bdev);
         dev->users++;
         spin_unlock(&dev->lock);
         return 0;
 }
 
 static int sbull_release(struct inode *inode, struct file *filp)
 {
         struct sbull_dev *dev = inode->i_bdev->bd_disk->private_data;
 
         spin_lock(&dev->lock);
         dev->users--;
 
         if (!dev->users) {
                 
         }
         spin_unlock(&dev->lock);
 
         return 0;
 }
 
 /*
  * Look for a (simulated) media change.
  */
 int sbull_media_changed(struct gendisk *gd)
 {
         return 1;
 }
 
 /*
  * Revalidate.  WE DO NOT TAKE THE LOCK HERE, for fear of deadlocking
  * with open.  That needs to be reevaluated.
  */
 int sbull_revalidate(struct gendisk *gd)
 {         
         return 0;
 }
 
 /*
  * The "invalidate" function runs out of the device timer; it sets
  * a flag to simulate the removal of the media.
  */
 void sbull_invalidate(unsigned long ldev)
 {
         
 }
 
 /*
  * The ioctl() implementation
  */
 
 int sbull_ioctl (struct inode *inode, struct file *filp,
                  unsigned int cmd, unsigned long arg)
 {
         return -ENOTTY; /* unknown command */
 }
 
 
 
 /*
  * The device operations structure.
  */
 static struct block_device_operations sbull_ops = {
         .owner           = THIS_MODULE,
         .open            = sbull_open,
         .release         = sbull_release,
         .ioctl           = sbull_ioctl
 };
 
 /*
  * Set up our internal device.
  */
 static void setup_device(struct sbull_dev *dev, int which)
 {
	memset (dev, 0, sizeof (struct sbull_dev));

	// -- set up checksum parameters
	dev->block_per_chksm = 2;
	dev->chksm_func = 2;
	dev->chksm_size = KERNEL_SECTOR_SIZE / dev->block_per_chksm;

	// Set up RAID parameters
	dev->layout_row = 4;
	dev->layout_column = 4;
	dev->parity_row = 4;
	dev->parity_column = 12;


	//Set up raid layout
	dev->raid_layout[0][0] = 1;
	dev->raid_layout[0][1] = 1;
	dev->raid_layout[0][2] = 1;
	dev->raid_layout[0][3] = 0;
	dev->raid_layout[1][0] = 1;
	dev->raid_layout[1][1] = 1;
	dev->raid_layout[1][2] = 0;
	dev->raid_layout[1][3] = 1;
	dev->raid_layout[2][0] = 1;
	dev->raid_layout[2][1] = 0;
	dev->raid_layout[2][2] = 1;
	dev->raid_layout[2][3] = 1;
	dev->raid_layout[3][0] = 0;
	dev->raid_layout[3][1] = 1;
	dev->raid_layout[3][2] = 1;
	dev->raid_layout[3][3] = 1;

	//Set up raid parity
	dev->raid_parity[0][0] = 1;
	dev->raid_parity[0][1] = 1;
	dev->raid_parity[0][2] = 1;
	dev->raid_parity[0][3] = 0;
	dev->raid_parity[0][4] = 0;
	dev->raid_parity[0][5] = 0;
	dev->raid_parity[0][6] = 0;
	dev->raid_parity[0][7] = 0;
	dev->raid_parity[0][8] = 0;
	dev->raid_parity[0][9] = 0;
	dev->raid_parity[0][10] = 0;
	dev->raid_parity[0][11] = 0;

	dev->raid_parity[1][0] = 0;
	dev->raid_parity[1][1] = 0;
	dev->raid_parity[1][2] = 0;
	dev->raid_parity[1][3] = 1;
	dev->raid_parity[1][4] = 1;
	dev->raid_parity[1][5] = 1;
	dev->raid_parity[1][6] = 0;
	dev->raid_parity[1][7] = 0;
	dev->raid_parity[1][8] = 0;
	dev->raid_parity[1][9] = 0;
	dev->raid_parity[1][10] = 0;
	dev->raid_parity[1][11] = 0;

	dev->raid_parity[2][0] = 0;
	dev->raid_parity[2][1] = 0;
	dev->raid_parity[2][2] = 0;
	dev->raid_parity[2][3] = 0;
	dev->raid_parity[2][4] = 0;
	dev->raid_parity[2][5] = 0;
	dev->raid_parity[2][6] = 1;
	dev->raid_parity[2][7] = 1;
	dev->raid_parity[2][8] = 1;
	dev->raid_parity[2][9] = 0;
	dev->raid_parity[2][10] = 0;
	dev->raid_parity[2][11] = 0;

	dev->raid_parity[3][0] = 0;
	dev->raid_parity[3][1] = 0;
	dev->raid_parity[3][2] = 0;
	dev->raid_parity[3][3] = 0;
	dev->raid_parity[3][4] = 0;
	dev->raid_parity[3][5] = 0;
	dev->raid_parity[3][6] = 0;
	dev->raid_parity[3][7] = 0;
	dev->raid_parity[3][8] = 0;
	dev->raid_parity[3][9] = 1;
	dev->raid_parity[3][10] = 1;
	dev->raid_parity[3][11] = 1;	

	int i=0, j=0, count = 0;

	for(i=0;i<dev->layout_row;i++){
		for(j=0;j<dev->layout_column;j++)
			if(dev->raid_layout[i][j] == 1)
				count++;
	}

	dev->data_per_stripe = count;
	dev->parity_per_stripe = dev->parity_row * dev->parity_column - dev->data_per_stripe;
	
        dev->tempU8 = vmalloc(KERNEL_SECTOR_SIZE);

        if (dev->tempU8 == NULL) {
                printk (KERN_NOTICE "vmalloc failure.\n");
                return -1;
        }

        memset(dev->tempU8,0, KERNEL_SECTOR_SIZE);
	

	for(i=0;i<11;i++){
		dev->temp[i] = vmalloc(KERNEL_SECTOR_SIZE);
		if(!dev->temp[i])
			goto out_vfree;
		memset(dev->temp[i], 0, KERNEL_SECTOR_SIZE);
	}
	for (i = 0; i < 3; i++) {
		dev->tempChksm[i] = vmalloc(dev->chksm_size);
		if (!dev->tempChksm[i])
			goto out_vfree;
		memset(dev->tempChksm[i], 0, dev->chksm_size);
	}
		
	/*
          * Get some memory.
          */
         
         dev->size = nsectors*hardsect_size;

	
	for(i=0;i<dev->layout_column;i++){
         	dev->data[i] = vmalloc(dev->size);

         	if (dev->data[i] == NULL) {
         	        printk (KERN_NOTICE "vmalloc failure.\n");
         	        return;
         	}
		memset(dev->data[i],0,dev->size);
	}
         spin_lock_init(&dev->lock);
         
                  
         /*
          * The I/O queue, depending on whether we are using our own
          * make_request function or not.
          */
         switch (request_mode) {
             case RM_NOQUEUE:
                 dev->queue = blk_alloc_queue(GFP_KERNEL);
                 if (dev->queue == NULL)
                         goto out_vfree;
                 blk_queue_make_request(dev->queue, sbull_make_request);
                 break;
 
             case RM_FULL:
                 dev->queue = blk_init_queue(sbull_full_request, &dev->lock);
                 if (dev->queue == NULL)
                         goto out_vfree;
                 break;
 
             default:
                 printk(KERN_NOTICE "Bad request mode %d, using simple\n", request_mode);
                 /* fall into.. */
         
             case RM_SIMPLE:
                 dev->queue = blk_init_queue(sbull_request, &dev->lock);
                 if (dev->queue == NULL)
                         goto out_vfree;
                 break;
         }
         blk_queue_hardsect_size(dev->queue, hardsect_size);
         dev->queue->queuedata = dev;
         /*
          * And the gendisk structure.
          */
         dev->gd = alloc_disk(SBULL_MINORS);
         if (! dev->gd) {
                 printk (KERN_NOTICE "alloc_disk failure\n");
                 goto out_vfree;
         }
         dev->gd->major = sbull_major;
         dev->gd->first_minor = which*SBULL_MINORS;
         dev->gd->fops = &sbull_ops;
         dev->gd->queue = dev->queue;
         dev->gd->private_data = dev;
         snprintf (dev->gd->disk_name, 32, "sbull%c", which + 'a');
         set_capacity(dev->gd, nsectors*(hardsect_size/KERNEL_SECTOR_SIZE));
         add_disk(dev->gd);
         return;
 
   out_vfree:{
	        int j;
	        for(j=0;j<dev->layout_column;j++){
			if(dev->data[j])
		         	vfree(dev->data[j]);
		}
		for(j=0;j<11;j++){
			if(dev->temp[j])
		         	vfree(dev->temp[j]);
		}
		for (j = 0; j < 3; j++) {
			if (dev->tempChksm[j])
				vfree(dev->tempChksm[j]);
		}
		if (dev->tempU8)
			vfree(dev->tempU8);
		
	}
 }
 
 
 
 static int __init sbull_init(void)
 {
         int i;
         /*
          * Get registered.
          */
         sbull_major = register_blkdev(sbull_major, "sbull");
         if (sbull_major <= 0) {
                 printk(KERN_WARNING "sbull: unable to get major number\n");
                 return -EBUSY;
         }
         /*
          * Allocate the device array, and initialize each one.
          */
         Devices = kmalloc(ndevices*sizeof (struct sbull_dev), GFP_KERNEL);
         if (Devices == NULL)
                 goto out_unregister;
         for (i = 0; i < ndevices; i++) 
                 setup_device(Devices + i, i);
     
         return 0;
 
   out_unregister:
         unregister_blkdev(sbull_major, "sbd");
         return -ENOMEM;
 }
 
 static void sbull_exit(void)
 {
         int i;
 
         for (i = 0; i < ndevices; i++) {
                 struct sbull_dev *dev = Devices + i;
 
                 if (dev->gd) {
                         del_gendisk(dev->gd);
                         put_disk(dev->gd);
                 }
                 if (dev->queue) {
                         if (request_mode == RM_NOQUEUE)
                                 blk_put_queue(dev->queue);
                         else
                                 blk_cleanup_queue(dev->queue);
                 }
		int j;
                for(j=0;j<dev->layout_column;j++){
			if(dev->data[j])
                        	vfree(dev->data[j]);
		}
		for(j=0;j<11;j++){
			if(dev->temp[j])
		         	vfree(dev->temp[j]);
		}
		for (j = 0; j < 3; j++) {
			if (dev->tempChksm[j])
				vfree(dev->tempChksm[j]);
                if (dev->tempU8)
                        vfree(dev->tempU8);

		}
         }
         unregister_blkdev(sbull_major, "sbull");
         kfree(Devices);
 }
         
 module_init(sbull_init);
 module_exit(sbull_exit);



